/* 
 * Configure a fourtun interface.
 *
 * Written 2000 by Andi Kleen, SuSE Labs.
 */ 
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <arpa/inet.h>

#include <asm/types.h>
#include <linux/sockios.h>
#include "fourtun.h"

#define err(x) perror(x),exit(1) 

int sock; 
char *cmdname; 

static struct in6_addr nulladdr; 
#define is_null_addr(adr) (!memcmp(&nulladdr, &(adr), sizeof(struct in6_addr)))

void fourstat(void) 
{ 
	struct ifreq ifr; 
	struct fourtun_stat fstat;
	strcpy(ifr.ifr_name,"fourtun0");  // XXX
	ifr.ifr_ifru.ifru_data = (char *)&fstat;  
	if (ioctl(sock, SIOCGTUNNEL6STAT, &ifr) < 0) { 
		err("ioctl SIOCGTUNNEL6STAT for fourtun0"); 
	} 
	printf("fourtun statistics:\n");
#define pstat(x) printf("\t%s: %u\n", #x, fstat.x)
#define pstat2(x,str) printf("\t%s: %u\n", str, fstat.x)
  	pstat2(unmatched_packets, "v4inv6 packets not matching any tunnel");
	pstat2(oom, "out of memory");
	pstat2(frags_needed, "fragmentation required"); 
	pstat2(pmtu_sent, "icmp too big sent");
	pstat2(pmtu_rcv, "icmpv6 packet too big received");
	pstat(pmtu_bad);
	pstat(pmtu_rmt_sent);
	pstat(pmtu_local_sent);
	pstat(unmatched_icmp);
	pstat(bad_icmp);
	pstat(sent_icmp);
	pstat(v6_in_v4);
	pstat(tunnel_loop);
	pstat(bad_pkt);
	pstat(mtu_change);
#undef pstat
#undef pstat2
	exit(0); 
} 

void printargs(char *dev, struct fourtun_args *args) 
{
	char src[32],dst[32]; 
	printf("%s:\n", dev); 
	inet_ntop(AF_INET6, &args->saddr, src, sizeof(src));
	inet_ntop(AF_INET6, &args->daddr, dst, sizeof(dst));
	printf("saddr:%s daddr:%s\n", src, dst); 
	printf("version:%u priority:%u hoplimit:%u encaplimit:%u link:%u\n",
		   args->version, args->priority, args->hoplimit, args->encaplimit,
		   args->link);
}

void usage(void)
{ 
	fprintf(stderr, "usage:\n%s [add|del|chg] device \n"
			"\t[priority pri] [hoplimit ttl] [encaplimit limit]\n"
			"\t[saddr saddr] [daddr daddr]\n", cmdname);
	fprintf(stderr, "%s show device\n", cmdname);
	fprintf(stderr, "%s stat\n", cmdname);
	exit(1); 
} 

void resolve_name(char *name, struct in6_addr *ip6)
{ 
	struct addrinfo ai, *res; 
	int err; 
	memset(&ai,0,sizeof(struct addrinfo)); 
	ai.ai_family = AF_INET6; 
	ai.ai_protocol = IPPROTO_IPV6; 
	err = getaddrinfo(name, NULL, NULL, &res);
	if (err) {
		fprintf(stderr,"%s: cannot find %s: %s\n", cmdname, name, gai_strerror(err));
		exit(1);
	}
	*ip6 = ((struct sockaddr_in6 *) (res->ai_addr))->sin6_addr;
	freeaddrinfo(res); 
} 

int main(int ac, char **av) 
{ 
	struct ifreq ifr; 
	int cmd; 
	struct fourtun_args f; 
	char **arg; 
	
	cmdname = av[0]; 
	memset(&f, 0, sizeof(struct fourtun_args));

	sock = socket(PF_INET,SOCK_DGRAM,0);
	if (sock < 0) 
		err("socket"); 

	if (ac < 2) { 
		usage(); 
	} 

	if (!strcmp(av[1],"stat")) {
		fourstat(); 
		exit(0); 
	} 

	ifr.ifr_ifru.ifru_data = (char *) &f; 
	
	arg = av + 1; 
	if (!strcmp(*arg, "add")) { 
		cmd = SIOCADD6TUNNEL; 
		++arg; 
	} else if (!strcmp(*arg, "del")) { 
		cmd = SIOCDEL6TUNNEL; 
		++arg; 
	} else if (!strcmp(*arg, "chg") || !strcmp(*arg,"change")) { 
		cmd = SIOCCHG6TUNNEL; 
		++arg; 
	} else if (!strcmp(*arg, "show")) { 
		cmd = SIOCGET6TUNNEL; 
		++arg;
		// XXX: read SIOCGIFCONF if no device specified
	} else { 
		cmd = SIOCADD6TUNNEL; 
	} 

	if (!*arg) 
		usage(); 

	strncpy(ifr.ifr_name, *arg, IFNAMSIZ); 
	arg++;

	f.version = 1;

	if (ioctl(sock, SIOCGET6TUNNEL, &ifr) < 0) { 
		f.priority = 0; 
		f.hoplimit = 64;
		f.encaplimit = 4; 
	}  

	for (;*arg; ++arg) { 
		if (!strcmp(*arg, "add")) { 
			cmd = SIOCADD6TUNNEL; 
			continue;
		} else if (!strcmp(*arg, "del")) { 
			cmd = SIOCDEL6TUNNEL; 
			continue; 
		} else if (!strcmp(*arg,"show")) { 
			cmd = SIOCGET6TUNNEL;
			continue;
		} else if (!strcmp(*arg, "chg") || !strcmp(*arg,"change")) { 
			cmd = SIOCCHG6TUNNEL; 
			continue;
		}

		if (cmd == SIOCGET6TUNNEL) 
			usage(); 
			
		if (!arg[1]) 
			usage(); 
		if (!strcmp(*arg, "priority")) { 
			f.priority = strtoul(arg[1],0,0); 
		} else if (!strcmp(*arg, "hoplimit"))
			f.hoplimit = strtoul(arg[1],0,0); 
		else if (!strcmp(*arg, "encaplimit"))
			f.encaplimit = strtoul(arg[1],0,0);
		else if (!strcmp(*arg, "saddr")) { 
		    resolve_name(arg[1], &f.saddr); 
		} else if (!strcmp(*arg, "daddr")) { 
			resolve_name(arg[1], &f.daddr); 
		} else if (!strcmp(*arg, "dev")) { 
			struct ifreq ifdev;
			strncpy(ifdev.ifr_name, arg[1], IFNAMSIZ); 
			if (ioctl(sock, SIOCGIFINDEX, &ifdev) < 0) { 
				fprintf(stderr,
						"%s: unknown device %s: %s\n", 
						cmdname, arg[1], strerror(errno)); 
				exit(1); 
			} 
			f.link = ifdev.ifr_ifindex; 
		} else { 
			usage();
		} 
		++arg;
	} 

	if (cmd == SIOCADD6TUNNEL) { 
		if (is_null_addr(f.saddr)) { 
			fprintf(stderr, "%s: saddr required\n", cmdname);
			exit(1); 
		}
		if (is_null_addr(f.daddr)) { 
			fprintf(stderr, "%s: daddr required\n", cmdname);
			exit(1); 
		}
	} 

	if (ioctl(sock, cmd, &ifr) < 0) 
		err("ioctl"); 

	if (cmd == SIOCGET6TUNNEL) { 
		printargs(ifr.ifr_name, &f); 
	} 

	exit(0); 	
} 
